#open "genlex";;
#open "printf";;

let keywords =
  ["variable"; "action"; "function"; "changed"; "invariant";
   ":"; ";"; "."; "?"]
;;

let lexer = make_lexer keywords;;

let repeat_parser f s =
  while (try end_of_stream s;false with Parse_failure -> true) do
    f s
  done
;;

(***********
let print_token =
  function
    [< 'Kwd k >] -> printf "KEY %s\n" k; ()
  | [< 'Ident i >] -> printf "IDENT %s\n" i; ()
  | [< 'Int i >] -> printf "INT %d\n" i; ()
  | [< '_ >] -> printf "BUG\n"; ()
;;
repeat_parser print_token (lexer (stream_of_channel stdin));;
flush stdout;;
************)

exception Multiple_decl of string
and Too_many_actions
and Unknown_var of string
;;

type id = { interf : int vect; mutable vars : string list };;

let id_table = (hashtbl__new 253 : (string, id) hashtbl__t);;

let maxact = 100;;

let new_ident i =
  try
    hashtbl__find id_table i;
    raise (Multiple_decl i)
  with
    Not_found -> hashtbl__add id_table i {interf=make_vect maxact 0;vars=[]}
;;
let get_ident i =
  try hashtbl__find id_table i
  with Not_found -> raise (Unknown_var i)
;;

let rec parse_variables s =
  match s with
    [< 'Ident i >] -> new_ident i; (get_ident i).vars <- [i]; parse_variables s
  | [< 'Kwd "." >] -> ()
;;

let var_in_act actnum var =
  if actnum >= maxact then raise Too_many_actions;
  (get_ident var).interf.(actnum) <- 2
;;

let rec parse_actions s =
  match s with
    [< 'Int n; 'Kwd ":" >] -> parse_act_vars n s
  | [< 'Kwd "." >] -> ()
and parse_act_vars n s =
  match s with
    [< 'Ident v >] -> var_in_act n v; parse_act_vars n s
  | [< 'Kwd ";" >] -> parse_actions s
;;

let merge_vect v w =
  for i = 0 to vect_length v - 1 do
    if w.(i) > 0 then
      v.(i) <- 1
  done
;;

let rec merge_lists l1 l2 =
  match (l1, l2) with
    ([], _) -> l2
  | (_, []) -> l1
  | (h1::t1, h2::t2)
    -> match compare_strings h1 h2 with
         0 -> h1 :: merge_lists t1 t2
       | (1|2) -> h2 :: merge_lists l1 t2
       | (-1|-2) -> h1 :: merge_lists t1 l2
       | _ -> failwith "compare_string"
;;

let var_in_fun f id =
  let v = get_ident id in
    merge_vect f.interf v.interf;
    f.vars <- merge_lists f.vars v.vars
;;

let rec parse_functions s =
  match s with
    [< 'Ident f; 'Kwd ":" >] -> new_ident f; parse_fun_vars (get_ident f) s
  | [< 'Kwd "." >] -> ()
and parse_fun_vars f s =
  match s with
    [< 'Ident i >] -> var_in_fun f i; parse_fun_vars f s
  | [< 'Kwd ";" >] -> parse_functions s
;;

let act_in_changed f a = f.interf.(a) <- 2;;

let all_in_changed f =
  for i = 0 to vect_length f.interf - 1 do
    if f.interf.(i) > 0 then f.interf.(i) <- 2
  done
;;

let rec parse_changeds s =
  match s with
    [< 'Ident f; 'Kwd ":" >] -> parse_changed_acts (get_ident f) s
  | [< 'Kwd "." >] -> ()
and parse_changed_acts f s =
  match s with
    [< 'Int a >] -> act_in_changed f a; parse_changed_acts f s
  | [< 'Kwd "?" >] -> all_in_changed f; parse_changed_acts f s
  | [< 'Kwd ";" >] -> parse_changeds s
;;

let invs = make_vect maxact ([| |] : int vect);;
for i = 0 to 99 do invs.(i) <- make_vect maxact 0 done;;

let max_vect v w =
  for i = 0 to vect_length v - 1 do
    if w.(i) > v.(i) then v.(i) <- w.(i)
  done
;;

let id_in_inv n id =
  let v = get_ident id in
  max_vect invs.(n) v.interf
;;

let rec parse_invariants s =
  match s with
    [< 'Int n; 'Kwd ":" >] -> parse_invariant_vars n s
  | [< 'Kwd "." >] -> ()
and parse_invariant_vars n s =
  match s with
    [< 'Ident i >] -> id_in_inv n i; parse_invariant_vars n s
  | [< 'Kwd ";" >] -> parse_invariants s
;;

let parse_phrase s =
  match s with
    [< 'Kwd "variable" >] -> parse_variables s
  | [< 'Kwd "action" >] -> parse_actions s
  | [< 'Kwd "function" >] -> parse_functions s
  | [< 'Kwd "changed" >] -> parse_changeds s
  | [< 'Kwd "invariant" >] -> parse_invariants s
;;

try
  repeat_parser parse_phrase (lexer (stream_of_channel stdin))
with
  Multiple_decl v -> fprintf stderr "Multiple declaration of %s\n" v
| Too_many_actions -> fprintf stderr "There are too many actions\n"
| Unknown_var v -> fprintf stderr "Unknown variable %s\n" v
| _ -> fprintf stderr "Unexpected exception\n"
;;

(*********
printf "variables:\n";;
let print_var name {interf=it; vars=v} =
  printf "%s " name;
  for i = 0 to maxact-1 do
    if it.(i) = 1 then
      printf "(%d) " i
    else if it.(i) = 2 then
      printf "%d " i
  done;
  printf "\n-> ";
  do_list (fun id -> printf "%s " id) v;
  printf "\n"
;;
hashtbl__do_table print_var id_table;;

printf "\n\ninvariants:\n";;
for i = 1 to 46 do
  printf "I%d " i;
  for j = 0 to maxact-1 do
    if invs.(i).(j) = 1 then
      printf "(%d) " j
    else if invs.(i).(j) = 2 then
      printf "%d " j
  done;
  printf "\n"
done;;
*************)

let ne = ref 0
and nf = ref 0
and nh = ref 0
;;

for i = 1 to 63 do
  printf "$a_{%d}$" i;
  for j = 1 to 46 do
    match invs.(j).(i) with
      0 -> printf "&\\e"; incr ne
    | 1 -> printf "&\\f"; incr nf
    | 2 -> printf "&\\h"; incr nh
    | _ -> fprintf stderr "***BUG***\n"
  done;
  printf "\\\\\n";
  if i mod 5 = 0 then printf "\\hline\n"
done;
printf "\\hline\n";
printf "%% number of \\e: %d; of \\f: %d; of \\h: %d\n" !ne !nf !nh
;;

flush stdout;;
flush stderr;;
